#ifndef CLUSTER_H
#define CLUSTER_H
#include "split.h"
#include <string.h>
#include <set>
#include <stdlib.h>
#include <stdio.h>
#include <cmath>
#include <iostream>
#include <vector>
#include <map>

using namespace std;

struct Pair
{
    int cluster1;
    int cluster2;
};

bool operator< (const Pair& a, const Pair& b);


class LmerClusters{
    private: 
        int n_of_lmers;
        string f_clusters;
        int * lmer_clusters;
        int * read_clusters;
        int n_clusters;
        int n_of_clustered_lmers;
        string f_clusters_freq;
        string ab_reads_file;
        int * read_bin;
        int abin;
        int n_reads;
        string reads_file;
        string reads_file_main;
        multiset <Pair> paired_connections;
        map<int,int> lmer_cluster_size;
        int min_cluster_size;
        Pair * top_connections;
        int min_pairs_connecting_clusters;
        set <Pair> pairs_to_merge;
        map<int,int> group_of_clusters;
        int n_of_groups; 
        set <int> * big_cluster;
        set <int> * reads_in_cluster;
        int ** m_graph;
        string mcl_input;
        int n_mcl_clusters;
        string output_prefix;
        int * clust_to_mclust;
    public:  
        LmerClusters(int _n_of_lmers, string _f_clusters, 
            string  _f_clusters_freq, string _ab_reads_file, int _abin, 
            string _reads_file, int _n_reads, string _reads_file_main, 
            string _mcl_input);
        LmerClusters(int _n_of_lmers, string _ab_reads_file,
            int _abin, string _reads_file, int _n_reads, 
            string _reads_file_main, string _output_prefix);
        void fillClusters(string fin);
        void outputStatistics(int interval_length);
        void initReadBin();
        void fillBin();
        int getNextIndex(int ind);
        void addOrderedPairedConnection(int clust1, int clust2);
        void addPairedConnection(int first_ind, int second_ind);
        void fillPairedConnections();
        void initReadClusters();
        int fillReadClusters();
        void connectClusters(string fout);
        int countPair(int clust_i, int clust_j);
        int top(int topj, int i);
        Pair topConnections(int i);
        void topConnections();
        bool candidatePair(int i, int j);
        void merge(int i,int j);
        void prepareForMerging();
        bool inGroup(int cluster1,int cluster2,  set <int> &group);
        void fillGroup(set <int> &group);
        void mapGroup(int group_number, set <int> &group);
        void mergeClusters(); 
        void processGroup(int &added, int cluster1, int cluster2, 
           set <int> &group);
        void recomputeClusters();
        void outputClusters2(string fout);
        void initBigClusters(); 
        void updateWithReads(int read_index_1, int read_index_2);
        void updateWithRead(int read_index);
        void updateReadClusters();
        void update(int cluster, set<int> & lmers);
        void updateBigClusters(set<int> & lmers, int read_num);
        void fillBigClusters();
        void buildMGraph(string fin, string fout);
        void initMGraph();
        void fillMGraph();
        void printMGraph();
        int intersect(int clust_i, int clust_j);
        int intersection(int clust_i, int clust_j); 
        void printClusters(string fout);    
        void fillMClusters(string fin);
        void updateClustersToMClusters();
        void updateReadsClusters();
        void updateLmersClusters();
        void separateReads();
        void finalReadsClustering(string fin,string mcl_fin); 
};



#endif
