#ifndef LMERGRAPH_H
#define LMERGRAPH_H

#include <string.h>
#include <set>
#include <stdlib.h>
#include <stdio.h>
#include <cmath>
#include <iostream>
#include <vector>
using namespace std;

class LmerGraph{
    private: 
        int n_of_nodes;
        int n_of_lmers;
        int n_of_reads;
        int max_n_of_neighbors;
        int min_counts;
        int * lmer_to_node;           
        int * node_to_lmer;
        int ** neighbors; 
        bool * node_is_out;
        int * lmer_counts;
        int * node_neighbors;
        string fin_counts;
        string f_reads;
        int count_threshold; 
        int min_unassigned_lmers;
        int min_number_of_neighbors;
        int max_number_of_nodes_to_add;
        int * node_cluster;
        int processed_reads;
        int number_of_clusters;
        string output_lmer_clusters;
        int read_length;
    public:  
        LmerGraph(string _fin_counts, int _n_of_lmers, int _max_n_of_neighbors,
            string _f_reads, int _count_threshold, int _min_unassigned_lmers, 
            int _min_number_of_neighbors, int _max_number_of_nodes_to_add, 
            string _output_lmer_clusters, int _min_counts, int _n_of_reads, 
            int _read_length);
        void fillLmers();                      
        void initGraph();                      
        void fillGraph();                     
        void refineGraph();                     
        void addClique(set<int> &clique);    
        bool isNeighbor(int lmer1,int lmer2);  
        bool isFull(int lmer1);                
        void addPair(int lmer1, int lmer2); 
        void fillNodes();                    
        void addNeighbor(int lmer1,int lmer2); 
        void phaseI();
        void candidateNodes(int node,multiset <int> &candidates_to_add,
        set<int> &nodes_just_added,set<int> &nodes_in_cluster);
        void newNodes(multiset <int> &candidates_to_add,set <int> &nodes_to_add);
        void removeFromCandidates(int node);
        void stopCluster(set <int> &nodes_to_add, set<int> &nodes_just_added);
        int updateCluster(set<int> &nodes_to_add, set<int> &nodes_just_added, 
            set<int> &nodes_in_cluster);
        int iteration(set<int> &nodes_in_cluster, set<int> &nodes_just_added);
        void addNewCluster(set <int> &nodes_in_cluster, int current_cluster);
        void growingClsuter(set <int> &initialized_cluster);
        bool candidateNode(int node);
        bool findUnassignedRead(set <int> &initialized_cluster, ifstream &file_reads);
        void outputClusters();
        void show_graph();
        void printProgress(int &reads_counter,double &persent);
        void outputLmerToNode();
        void outputNodeToLmer();
        bool validNextParameters();
        void nextParameters();
        void resetParameters();
};



#endif
