#include "MainProfile.h"
#include "LmerGraph.h"
#include "Lambda.h"
#include "Cluster.h"
#include "split.h"
#include <iostream>
#include <cmath>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>
#include <sys/stat.h>
#include <sys/types.h>
#define MY_MASK S_IRWXU | S_IROTH | S_IRGRP | S_IXOTH | S_IXGRP

int MainProfile::reads_counter = 0; 
int MainProfile::n_unique = 0;
int MainProfile::n_all = 0;
int MainProfile::l = 20;
string MainProfile::input_fasta_file = "";
string MainProfile::dir = "my_dir";
string MainProfile::tmp_dir = "tmp";
string MainProfile::main_tmp_dir = "tmp";
string MainProfile::abundancebin_dir_prefix = "ab";
int MainProfile::n_of_bins_abundancebin = 0;
string MainProfile::abundancebin_output_prefix = "";
int *MainProfile::n_ab_reads = NULL;
int *MainProfile::n_unique_in_bin = NULL;
int MainProfile::n_reads = 0;
int MainProfile::read_length = 0;

void checkOrMakeDir(string dir_name)
{
    if (access(dir_name.c_str(),0))
    {
          int temp;
          temp = umask(0);
          if ((temp = mkdir(dir_name.c_str(), MY_MASK)) != 0) 
    		cout << "Error: unable to mkdir " << dir_name << endl;
    }
}

char nucleotideNum(char a)
{  
    if (a == 'A' || a == 'a') return '0';
        else if (a == 'C' || a == 'c') return '1';
            else if (a == 'T' || a == 't') return '3';
                else if (a == 'G' || a == 'g') return '2'; 
    return 'x'; 
}

int64_t numericToNum(string lmer)
{
    int64_t sum = 0;
    int order = lmer.length();
    for(int i = 0; i < order; i++)
	sum = sum + atoi(lmer.substr(i,1).c_str()) * pow(4,i);
    return sum;    
}

int64_t numericToNumReversed(string lmer)
{
    int64_t sum = 0;
    int order = lmer.length(); 
    for(int i = 0; i < order; i++)
	sum = sum + (3 - atoi(lmer.substr(order - i - 1,1).c_str())) * pow(4,i); 
    return sum;    
}

int64_t min(int64_t lmer_numeric_1, int64_t lmer_numeric_2)
{
    if (lmer_numeric_1 < lmer_numeric_2) 
        return lmer_numeric_1;
    else 
        return lmer_numeric_2;
}

RawRead::RawRead()
{
    clear();
}

RawRead::~RawRead()
{
    clear();
}

RawRead::RawRead(string _seq)
{
    read_letter_seq = _seq;
    read_numeric_seq = seqToNumeric();
}

void RawRead::append(string _seq)
{
    read_letter_seq += _seq;
    read_numeric_seq += seqToNumeric(_seq);
}

string RawRead::seqToNumeric()
{ 
    return seqToNumeric(read_letter_seq);       
}

string RawRead::seqToNumeric(string seq)
{ 
    string temp = ""; 
    for (int i = 0; i < seq.length(); i++)
        if (seq[i] != 'x') temp += nucleotideNum(seq[i]);
    return temp;     
}

string RawRead::getNumeric()
{ 
    return read_numeric_seq;    
}

void RawRead::clear()
{ 
    read_letter_seq = "";
    read_numeric_seq = "";     
}

MainProfile::MainProfile(int _l, string _input_fasta_file, string _dir,
    string _abundancebin_output_prefix, int _n_of_bins_abundancebin)
{
    l = _l;
    input_fasta_file = _input_fasta_file;
    dir =  _dir;
    tmp_dir = dir + "/tmp";
    stringstream i_out;
    n_of_bins_abundancebin = _n_of_bins_abundancebin;
    abundancebin_output_prefix = _abundancebin_output_prefix;
    if (n_of_bins_abundancebin > 0) 
    {
        for (int i = 0; i < n_of_bins_abundancebin; i++)
        {
            stringstream i_out;
            i_out << i + 1;
            string dir_name = dir +  "/" + abundancebin_dir_prefix + i_out.str();
            abundancebin_dir.push_back(dir_name)  ; 
        }
     }
     n_ab_reads = new int [n_of_bins_abundancebin];
     n_unique_in_bin = new int [n_of_bins_abundancebin];
     bad_bins = tmp_dir + "/bad_bins.txt";
}

MainProfile::MainProfile(string _dir, int _n_of_bins_abundancebin)
{    
    dir =  _dir;
    tmp_dir = dir + "/tmp";
    stringstream i_out;
    n_of_bins_abundancebin = _n_of_bins_abundancebin;
    if (n_of_bins_abundancebin > 0) 
    {
        for (int i = 0; i < n_of_bins_abundancebin; i++)
        {
            stringstream i_out;
            i_out << i + 1;
            string dir_name = dir +  "/" + abundancebin_dir_prefix + i_out.str();
            abundancebin_dir.push_back(dir_name)  ; 
        }
     }
     n_ab_reads = new int [n_of_bins_abundancebin];
     n_unique_in_bin = new int [n_of_bins_abundancebin];
     bad_bins = tmp_dir + "/bad_bins.txt";
}

void MainProfile::initHashedLmers(int num)
{
    if (num > 0)
    {
        hashed_lmers = new int64_t [num];
        for (int i = 0; i < num; i++)
            hashed_lmers[i] = 0;
    }
}
void MainProfile::initLmerCounts(int num)
{
    if (num > 0)
    {       
        lmer_counts = new unsigned int [num];
        for (int i = 0; i < num; i++)
            lmer_counts[i] = 0;
    }
}

void MainProfile::initReadBin(int num)
{
    if (num > 0)
    {       
        read_bin = new int [num];
        for (int i = 0; i < num; i++)
            read_bin[i] = -1;   //-1: unclustered by abundancebin
    }
}

void MainProfile::clearHashedLmers()
{    
    delete[] hashed_lmers;
}

void MainProfile::clearLmerCounts()
{    
    delete[] lmer_counts;
}

void MainProfile::clearReadBin()
{    
    delete[] read_bin;
}

void MainProfile::fillHash()
{ 
    string file_hashed_lmers = tmp_dir;
    file_hashed_lmers += "/hashed_lmers.txt";
    ifstream file_hash; 
    file_hash.open(file_hashed_lmers.c_str());
    string cur_s;
    while(getline(file_hash,cur_s))
    {
        vector<string> line = split(cur_s, ' ');
        int index = atoi(line[0].c_str());
        int64_t value = atoll(line[1].c_str());
        hashed_lmers[index - 1] = value;
    }
    file_hash.close();
} 

void MainProfile::fillLmers()
{    
    string file_lmers = tmp_dir;
    file_lmers += "/lmer_frequency.txt";
    ifstream file_lmers_freq; 
    file_lmers_freq.open(file_lmers.c_str());
    string cur_s;
    while(getline(file_lmers_freq,cur_s))
    {
        vector<string> line = split(cur_s, ' ');
        int index = atoi(line[0].c_str());
        int value = atoi(line[1].c_str()); 
        lmer_counts[index - 1] = value;
    }
    file_lmers_freq.close();        
}

int MainProfile::recursiveFindIndex(int64_t value,int lower_bound,
    int upper_bound)
{
    int index = -1;
    int median = trunc((upper_bound - lower_bound) / 2.0) + lower_bound;
    if (hashed_lmers[median] == value)
        return median;
    else if (lower_bound >= upper_bound)
        return -1;
    else 
    {
        if (hashed_lmers[median] > value) 
            index = recursiveFindIndex(value,lower_bound,median);
	else 
            index = recursiveFindIndex(value,median + 1,upper_bound);
    }
    return index;
}

int MainProfile::findIndex(int64_t value)
{
    return recursiveFindIndex(value,0,n_all - 1);
} 

void MainProfile::countFrequency()
{   
    string file_hashed_reads = tmp_dir;
    file_hashed_reads += "/hashed_reads.txt";
    ofstream file_hashed; 
    file_hashed.open(file_hashed_reads.c_str());

    string file_input_reads = tmp_dir;
    file_input_reads += "/input_reads.txt";
    ifstream file_input; 
    file_input.open(file_input_reads.c_str());
    string cur_s;
    int curr_len = 0;
    int max_len = 0;

    while(getline(file_input,cur_s))
    {
        if (cur_s[0]!='>')
        {
            curr_len ++;
            vector<string> line = split(cur_s, ' ');
            int64_t value = atoll(line[0].c_str());				
	    int index = findIndex(value);             
            if (index > -1)
            {
                lmer_counts[index] ++;
	        file_hashed << index << endl;                 
            }		
        }
        else 
        {
            file_hashed << cur_s << endl;
            if (curr_len > max_len) 
                max_len = curr_len;
            curr_len = 0;
        }
    }    
    file_hashed.close();
    file_input.close();
    read_length = max_len;
    remove(file_input_reads.c_str());
}

void MainProfile::countFrequency(int i)
{   
    string input_hashed_reads = tmp_dir;
    input_hashed_reads += "/hashed_reads.txt";
    ifstream file_input_hashed_reads; 
    file_input_hashed_reads.open(input_hashed_reads.c_str());

    string output_hashed_reads = abundancebin_dir[i];
    output_hashed_reads += "/hashed_reads.txt";
    ofstream file_output_hashed_reads; 
    file_output_hashed_reads.open(output_hashed_reads.c_str());

    string cur_s;
    reads_counter = -1;
    bool flag = false;
    int counter = 0;
    while(getline(file_input_hashed_reads,cur_s))
    {
        if (cur_s[0]=='>')      
        {
            reads_counter++;
            if (read_bin[reads_counter] == i )
                 //|| (reads_counter!=0 && read_bin[reads_counter-1] == i)) 
            {
                flag = true;
                counter++;
                file_output_hashed_reads << cur_s << endl;
            }	
            else 
                flag = false;
        }
        else if (flag)
        {
            file_output_hashed_reads << cur_s << endl; 
            vector<string> line = split(cur_s, ' ');
            int index = atoi(line[0].c_str());
            lmer_counts[index] ++;   
        }
    }  
    n_ab_reads[i] = counter; 
    file_output_hashed_reads.close();
    file_input_hashed_reads.close();
}

void MainProfile::fillBin(int i)
{
  //  cout << "Bin " << i+1 << "... "; 
    string i_str;
    stringstream i_out;
    i_out << i + 1;   
    string input_fname = "";
    input_fname += abundancebin_output_prefix + "." + i_out.str();
    ifstream file_input; 
    file_input.open(input_fname.c_str()); 
    string cur_s;
    while(getline(file_input,cur_s))
    {
        vector<string> line = split(cur_s, ' ');
        string read_name = line[0];
        int read_num = atoi(read_name.substr(4).c_str()) - 1;
        read_bin[read_num] = i;
    } 
       
    file_input.close();   
} 

int MainProfile::freqDistribution(string dir_name)
{
    string file_distribution = dir_name;
    file_distribution += "/lmer_frequency_distribution.txt";
    ofstream file_freq_distr; 
    file_freq_distr.open(file_distribution.c_str());

    string lmer_freq = dir_name;
    lmer_freq += "/lmer_frequency.txt";
    ofstream file_lmer_freq; 
    file_lmer_freq.open(lmer_freq.c_str());

    set<unsigned int> frequencies;
    multiset <unsigned int> frequencies_counts;
    set<unsigned int>::iterator it;
    int counter_of_all_lmers = 0;

    for (int i = 0; i < n_all; i++)
    {
        if (lmer_counts[i] > 1)
        {
            counter_of_all_lmers++ ;
        }
        file_lmer_freq << i + 1 << " " << lmer_counts[i] << endl;		
        frequencies_counts.insert(lmer_counts[i]);
        frequencies.insert(lmer_counts[i]);
    }

    int n_unique_lmers = 0, n_total_lmers = 0;

    for (it = frequencies.begin(); it != frequencies.end(); it++)
        file_freq_distr << *it << " " << frequencies_counts.count(*it) << " " 
            << frequencies_counts.count(*it) * (*it) << endl;
    file_lmer_freq.close();
    file_freq_distr.close(); 
    return counter_of_all_lmers;  
}



void MainProfile::outputN()
{
    string f_num_lmers = tmp_dir;
    f_num_lmers += "/number_of_lmers.txt";
    ofstream file_num_lmers; 
    file_num_lmers.open(f_num_lmers.c_str());
    file_num_lmers << n_unique << " " << n_all << endl << n_reads << " " 
        << read_length << endl;  
    file_num_lmers.close();
}

void MainProfile::outputN(int i)
{
    string f_num_lmers = abundancebin_dir[i];
    f_num_lmers += "/number_of_lmers.txt";
    ofstream file_num_lmers; 
    file_num_lmers.open(f_num_lmers.c_str());
    file_num_lmers << n_unique_in_bin[i] << endl << n_ab_reads[i] << endl;   
    file_num_lmers.close();
}


MainProfile::~MainProfile()
{  
    lmers.clear();  
}	

void MainProfile::processRawRead(RawRead &rd, ofstream &out_reads)
{
    if (rd.getNumeric().length() != 0)
    {
        reads_counter++;
        if (out_reads.is_open())
            out_reads << ">read " << reads_counter << endl;
        string lmer;
        int num_of_lmers = rd.getNumeric().length() - l + 1;
        for (int i = 0; i < num_of_lmers; i++)
        {
            lmer = rd.getNumeric().substr(i,l);
            int64_t lmer_as_num = numericToNum(lmer);
            int64_t lmer_as_num_reversed = numericToNumReversed(lmer);
            int64_t min_lmer_as_num = min(lmer_as_num,lmer_as_num_reversed);
       
            lmers.insert(min_lmer_as_num);
            if (out_reads.is_open())
                  out_reads << min_lmer_as_num << endl;                		                   
        }
    }
}

void MainProfile::listUniqueLmers()
{
    ifstream input_fasta;
    string cur_s;
    ofstream out_reads;
    string file_input_reads = tmp_dir;
    file_input_reads += "/input_reads.txt";
    out_reads.open(file_input_reads.c_str());
    input_fasta.open(input_fasta_file.c_str()); 
    RawRead rd;  

    while(getline(input_fasta,cur_s))
    {	
        if(cur_s[0]!='>')            
            rd.append(cur_s);        
	else
        {                
	    processRawRead(rd,out_reads);            
            rd.clear();          
        }
    }
    
    processRawRead(rd,out_reads); 
    n_reads = reads_counter;
    reads_counter = 0;    
    out_reads.close();
    input_fasta.close();
}

void MainProfile::buildHash()
{
    ofstream file_hash; 
    string file_hashed_lmers = tmp_dir;
    file_hashed_lmers += "/hashed_lmers.txt";
    file_hash.open(file_hashed_lmers.c_str());
    set<int64_t>::iterator it; 
    int counter_of_all_lmers = 0;

    for (it = lmers.begin(); it != lmers.end(); it++)
    {    
        counter_of_all_lmers++;		
        file_hash << counter_of_all_lmers << " " << *it << endl;
    } 

    file_hash.close();
    n_all = counter_of_all_lmers;
}

void MainProfile::checkDir()
{    
    checkOrMakeDir(dir);
    checkOrMakeDir(tmp_dir);
    if (n_of_bins_abundancebin > 0) 
        for (int i = 0; i < n_of_bins_abundancebin; i++)          
             checkOrMakeDir(abundancebin_dir[i]);   
}

void MainProfile::processingAbInput(int i)
{    
   // cout << "Hashing bin " << i+1 << "... "; 
    countFrequency(i); 
    n_unique_in_bin[i] = freqDistribution(abundancebin_dir[i]); 
    outputN(i);    
}

void MainProfile::fillReadsBins()
{
    if (n_of_bins_abundancebin > 0) 
        for (int i = 0; i < n_of_bins_abundancebin; i++)
            fillBin(i);
}

void MainProfile::hashLmers()
{     
    cout << "Pre-processing: Listing l-mers..." << endl;
    listUniqueLmers();
    cout << "Pre-processing: Hashing l-mers..." << endl;
    buildHash(); 
    cout << "Pre-processing: Loading hashed l-mers..." << endl;
    initHashedLmers(n_all);  
    fillHash(); 
    cout << "Pre-processing: Counting l-mers..." << endl;
    initLmerCounts(n_all);
    countFrequency(); 
    clearHashedLmers(); 
    n_unique = freqDistribution(tmp_dir); 
    clearLmerCounts();  
    cout << "Pre-processing is completed. Total number of distinct l-mers: " << n_all <<
        "; l-mers that occur more than once: " << n_unique << endl;
    outputN();
}

void MainProfile::initN()
{
    string f_num_lmers = tmp_dir;
    f_num_lmers += "/number_of_lmers.txt";
    ifstream file_num_lmers; 
    file_num_lmers.open(f_num_lmers.c_str());
    string s;
    getline(file_num_lmers,s);
    vector<string> line = split(s, ' ');
    n_unique = atoi(line[0].c_str());	
    n_all = atoi(line[1].c_str());
    getline(file_num_lmers,s);
    vector<string> line2 = split(s, ' '); 
    n_reads = atoi(line2[0].c_str());
    read_length = atoi(line2[1].c_str());
    file_num_lmers.close();   
}

void MainProfile::initN(int i)
{
    string f_num_lmers = abundancebin_dir[i];
    f_num_lmers += "/number_of_lmers.txt";
    ifstream file_num_lmers; 
    file_num_lmers.open(f_num_lmers.c_str());
    string s;
    getline(file_num_lmers,s);
    getline(file_num_lmers,s);
    vector<string> line2 = split(s, ' '); 
    n_ab_reads[i] = atoi(line2[0].c_str());
    file_num_lmers.close();   
}

void MainProfile::processingAbInput()
{   
    initN(); 
    cout << "----------------------------------------" << endl;
    cout << "Processing AbundanceBin files. Loading information about bins... " << endl; 
    initReadBin(n_reads);
    fillReadsBins();
    cout << "Processing AbundanceBin files. Hashing reads in the bins... " << endl; 
    if (n_of_bins_abundancebin > 0) 
        for (int i = 0; i < n_of_bins_abundancebin; i++)
        {
            initLmerCounts(n_all);
            processingAbInput(i);
            clearLmerCounts(); 
        }
    cout << "Processing of AbundanceBin files is completed." << endl;
    
    clearReadBin();
}

void MainProfile::clusteringBin(int i)
{   
    initN(); 
    initN(i); 

    cout << "----------------------------------------" << endl;

    cout << "Phase I (bin " << i + 1 << "). Estimation of thresholds." << endl;

    string fin_counts = abundancebin_dir[i] + "/lmer_frequency.txt";
    string fin_freq_distr = abundancebin_dir[i] + 
        "/lmer_frequency_distribution.txt";
    int n_of_lmers = n_all;

    FrequencyProfile fp(fin_counts, fin_freq_distr, n_of_lmers);
    fp.findThresholds();
    double lambda = fp.getLambda();
    
    cout << "Average l-mer frequency for bin " << i + 1 << " is " << lambda << endl;
    if (lambda > 5){
      int min_counts = fp.LowerT();
      int count_threshold = fp.UpperT();
      cout << "Lower count threshold: " << min_counts << "." << endl <<
          "Upper count threshold: " << count_threshold << endl; 
      int max_n_of_neighbors = read_length * 2 - 2;
      int min_unassigned_lmers = read_length / 2 + 1; 
      int min_number_of_neighbors = 1; 
      int max_number_of_nodes_to_add = 2 * (read_length - min_number_of_neighbors);
      string f_reads = abundancebin_dir[i] + "/hashed_reads.txt";
      string output_lmer_clusters = abundancebin_dir[i] + "/lmer_clusters.txt";
      int n_of_reads = n_ab_reads[i];

      LmerGraph lg(fin_counts, n_of_lmers, max_n_of_neighbors, f_reads, 
         count_threshold, min_unassigned_lmers, min_number_of_neighbors, 
         max_number_of_nodes_to_add, output_lmer_clusters, min_counts, 
         n_of_reads, read_length);
      lg.phaseI();
      cout << "Phase I (bin " << i + 1 << ") is completed." << endl;
    }
    else
    {
        cout << "Clustering of bin " << i + 1 << 
            " is not going to proceed. Insufficient coverage." << endl;
        unclustered_bins.insert(i);
    }    
}

void MainProfile::outputBadBins()
{
    ofstream fout;
    fout.open(bad_bins.c_str());
    for (set<int> :: iterator it = unclustered_bins.begin(); it != unclustered_bins.end(); it ++ )
        fout << *it << " ";
    fout.close();
}

void MainProfile::inputBadBins()
{
    unclustered_bins.clear();
    ifstream fin;
    fin.open(bad_bins.c_str());
    string cur_s;
    getline(fin,cur_s);
    vector<string> line = split(cur_s, ' ');
    for (int i = 0; i < line.size(); i++)        
        unclustered_bins.insert(atoi(line[i].c_str()));
    fin.close();
}

void MainProfile::clusterProcessing(int i)
{   
    if (!unclustered_bins.count(i))
    {
      cout << "----------------------------------------" << endl;
      initN(); 
      initN(i); 

      int n_of_lmers = n_all;
      string output_lmer_clusters = abundancebin_dir[i] + "/lmer_clusters.txt";
      string lmer_clusters_2 = abundancebin_dir[i] + "/lmer_clusters_2.txt";
      string f_clusters_freq = abundancebin_dir[i] + "/clusters_stat.txt";
      string i_str;
      stringstream i_out;
      i_out << i + 1;   
      string ab_reads_file = abundancebin_output_prefix + "." + i_out.str();
      string reads_file = abundancebin_dir[i] + "/hashed_reads.txt";
      string reads_file_main = tmp_dir + "/hashed_reads.txt";
      string mcl_input = abundancebin_dir[i] + "/mgraph.mci";
      string clust_info_file = abundancebin_dir[i] + "/cluster_info.txt";

      LmerClusters lc(n_of_lmers, output_lmer_clusters, f_clusters_freq, 
        ab_reads_file, i, reads_file, n_reads, reads_file_main, mcl_input);
 
//      cout << "Computing cluster statistics..." << endl;
//      lc.outputStatistics(500);
      
      cout << "Additional step to connect more clusters..." << endl;
      lc.connectClusters(lmer_clusters_2);
  
      cout << "Phase 2 (bin " << i + 1 << "). Building M-Graph..." << endl;
      lc.buildMGraph(lmer_clusters_2, clust_info_file);  
    }  
}

void MainProfile::finalReadsClustering(int i)
{   
    if (!unclustered_bins.count(i))
    {
      cout << "----------------------------------------" << endl;
      cout << "Post-processing (bin " << i + 1 << ")." << endl;
      initN(); 
      initN(i); 

      int n_of_lmers = n_all;
      string lmer_clusters_2 = abundancebin_dir[i] + "/lmer_clusters_2.txt";
      string i_str;
      stringstream i_out;
      i_out << i + 1;   
      string ab_reads_file = abundancebin_output_prefix + "." + i_out.str();
      string reads_file = abundancebin_dir[i] + "/hashed_reads.txt";
      string reads_file_main = tmp_dir + "/hashed_reads.txt";
      string mcl_fin = abundancebin_dir[i] + "/mclusters.txt";
      string output_prefix = dir + "/" + "reads." + i_out.str();
   
      LmerClusters lc(n_of_lmers,ab_reads_file, i, reads_file, n_reads, 
        reads_file_main,output_prefix);
      lc.finalReadsClustering(lmer_clusters_2,mcl_fin);
    }
}




