#ifndef MAINPROFILE_H
#define MAINPROFILE_H

#include <string.h>
#include <set>
#include <stdlib.h>
#include <stdio.h>
#include <cmath>
#include <iostream>
#include <vector>
using namespace std;

class RawRead{
    private: 
        string read_letter_seq;
        string read_numeric_seq;
    public:  
        RawRead(); 
        RawRead(string _seq);
        ~RawRead(); 
        void append(string s); 
        void convert(); 
        void clear();
        string seqToNumeric(string _seq);
        string seqToNumeric(); 
        string getNumeric(); 
};

class MainProfile{
    private: 
        static int reads_counter;
        set <int64_t> lmers;
        static int n_unique;
        static int n_all;
        static int n_reads;
        static int *n_ab_reads;
        static int *n_unique_in_bin;
        static int l;
        static string input_fasta_file;
        static string dir;
        static string tmp_dir;
        int64_t * hashed_lmers;
        unsigned int * lmer_counts;
        int * read_bin;
        static string main_tmp_dir;
        static string abundancebin_dir_prefix;
        static int n_of_bins_abundancebin;
        vector <string> abundancebin_dir;
        static string abundancebin_output_prefix;
        static int read_length;
        set <int> unclustered_bins;
        string bad_bins;
    public:  
        MainProfile(int _l, string _input_fasta_file, string _dir,  
            string _abundancebin_output_prefix, int _n_of_bins_abundancebin);
        MainProfile(string _dir, int _n_of_bins_abundancebin);
        ~MainProfile();
        void appendLmer(int64_t lmer); 
        void hashLmers();
        void processRawRead(RawRead &rd, ofstream &out_reads);
        void listUniqueLmers();
        void buildHash();
        void checkDir();
        void initHashedLmers(int num);
        void initLmerCounts(int num);
        void initReadBin(int num);
        void clearHashedLmers();
        void clearLmerCounts();
        void clearReadBin();
        void fillHash();
        void fillLmers();
        void countFrequency();
        void countFrequency(int i);
        int freqDistribution(string dir_name);
        int recursiveFindIndex(int64_t value,int lower_bound,int upper_bound);
        int findIndex(int64_t value);
        void outputN();
        void outputN(int i);
        void processingAbInput(int i);
        void processingAbInput();
        void fillReadsBins();
        void fillBin(int i); 
        void initN(); 
        void initN(int i); 
        void clusteringBin(int i);
        void clusterProcessing(int i); 
        void finalReadsClustering(int i); 
        void outputBadBins();
        void inputBadBins();
};

#endif
