#include "MclAnalyzer.h"
#include "split.h"
#include <iostream>
#include <cmath>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>

string MclOutput::mcl_path = "mcl";
double MclOutput::mcl_parameter_start_vlaue = 2.5;
int MclOutput::big_cluster_size = 400000;     
         

MclOutput::MclOutput(string _dir, int _bin, int _big_cluster_size)
{
    stringstream bin_num;
    bin_num << _bin;   
    abundancebin_dir_prefix = _dir +  "/ab" + bin_num.str() + "/";
    cluster_info = abundancebin_dir_prefix + "cluster_info.txt";
    mcl_output_fname = abundancebin_dir_prefix + "out.mgraph"; 
    mgraph_fname = abundancebin_dir_prefix + "mgraph.mci"; 
    final_mclusters = abundancebin_dir_prefix + "mclusters.txt"; 
    loadClustersInformation();
    big_cluster_size = _big_cluster_size;
}


void MclOutput::loadClustersInformation()
{
    ifstream clust_info_file;
    clust_info_file.open(cluster_info.c_str());
    string cur_s;  
    getline(clust_info_file,cur_s);
    vector<string> line = split(cur_s, ' ');
    n_clusters = atoi(line[0].c_str());          
    
    cluster_size = new int [n_clusters + 1];
    cluster_in_mcluster = new int [n_clusters + 1];
    for (int i = 1; i <= n_clusters; i++)
    {
        cluster_size[i] = 0;
        cluster_in_mcluster[i] = -1;
    }

    while(getline(clust_info_file,cur_s))
    {     
        vector<string> line2 = split(cur_s, ' ');
        int cluster = atoi(line2[0].c_str()); 
        int size = atoi(line2[1].c_str()); 
        cluster_size[cluster] = size;
    }
    clust_info_file.close(); 
}

void MclOutput::getNOfMclClusters()
{
    ifstream mcl_file;
    mcl_file.open(mcl_output_fname.c_str());
    string cur_s;  
    while(getline(mcl_file,cur_s) && cur_s[0]!='d') {}
    n_clusters_mcl = atoi((cur_s.substr(cur_s.find("x") + 1)).c_str());
    mcl_file.close(); 
}

void MclOutput::fillMGraph(vector<string> &line)
{
    int len = line.size();
    int cluster = atoi(line[0].c_str());
    for (int i = 1; i < len - 1; i++)
    {
        int cluster2 = atoi(line[i].substr(0,line[i].find(":")).c_str()); 
        int value = atoi(line[i].substr(line[i].find(":") + 1).c_str()); 
        mgraph[cluster][cluster2] = value;
    }
}

void MclOutput::fillMclCluster(vector<string> &line, int &m_cluster)
{
    int len = line.size();
    if (m_cluster == -1) m_cluster = atoi(line[0].c_str());
    int last = len ;
    if (line[len-1] == "$")
        last = len - 1;
    for (int i = 1; i < last; i++)
    {
        int clust = atoi(line[i].c_str());
        if (clust > 0)
        {
            mcl_cluster[m_cluster].insert(clust);           
            cluster_in_mcluster[clust] = m_cluster;
        }
    }
    if (line[len-1] == "$") m_cluster = -1;
}

void MclOutput::loadMGraph()
{
    mgraph = new int * [n_clusters + 1];

    for (int i = 0; i <=  n_clusters; i++)
        mgraph[i] = new int [n_clusters + 1];

    for (int i = 0; i < n_clusters + 1; i++)
        for (int j = 0; j < n_clusters + 1; j++)
            mgraph[i][j] = 0 ;


    ifstream mgraph_file;
    string cur_s;
    mgraph_file.open(mgraph_fname.c_str());

    while(getline(mgraph_file,cur_s) && cur_s[0]!='b') {}

    while(getline(mgraph_file,cur_s) && cur_s[0]!=')')
    {    
        vector<string> line = split(cur_s, ' ');
        fillMGraph(line);
    }
    mgraph_file.close();
 
   
    mcluster_graph = new  int *[n_clusters_mcl];

    for (int i = 0; i < n_clusters_mcl; i ++ )
        mcluster_graph[i] = new int [n_clusters_mcl];

    for (int i = 0; i < n_clusters_mcl; i++)
        for (int j = 0; j < n_clusters_mcl; j++)
            mcluster_graph[i][j] = 0 ;

    for (int i = 1; i <= n_clusters; i++)
    {
        int mclust1 = cluster_in_mcluster[i];
        for (int j = 1; j <= n_clusters; j++)
        {
            int mclust2 = cluster_in_mcluster[j]; 
            mcluster_graph[mclust1][mclust2] += mgraph[i][j];           
        }
    } 

}

void MclOutput::loadMclCLusters()
{
    for (int i = 1; i <= n_clusters; i++)
        cluster_in_mcluster[i] = -1;

    getNOfMclClusters();
    mcl_cluster = new set<int> [n_clusters_mcl];
    string cur_s;
    ifstream mcl_file;
    int m_cluster = -1;
    mcl_file.open(mcl_output_fname.c_str());

    while(getline(mcl_file,cur_s) && cur_s[0]!='b') {}
    while(getline(mcl_file,cur_s) && cur_s[0]!=')')
    {     
        vector<string> line = split(cur_s, ' ');
        fillMclCluster(line,m_cluster);
    }
    mcl_file.close();
   
}

int MclOutput::isBig(int mclust_i)
{
    int size = 0; 
    for (set <int> ::iterator it = mcl_cluster[mclust_i].begin(); 
        it != mcl_cluster[mclust_i].end(); it ++ )
        size += cluster_size[*it];   
    if (size >= big_cluster_size) 
        return 1;
    return 0;
}

int MclOutput::isGood(int i, int &contin)
{
    int good = 1;
    for (int j = 0; j < n_clusters_mcl; j ++)
    {       
        float statistics = exp(log(pow(10,6))  + log( mcluster_graph[i][j]) - 
            log (mcluster_graph[i][i]) -log( mcluster_graph[j][j]));      
        if (statistics > 0.95 && i != j)
            contin = 1; 
        if (i != j && isBig(j) && statistics > 0.95)
            good = 0;
    }  
    return good;      
}

void MclOutput::runMcl()
{
    std::stringstream ss;
    ss << mcl_parameter;
    string mcl_del = "rm " + mcl_output_fname;
    string mcl_run = mcl_path + " " + mgraph_fname + " -I " + 
        ss.str() + " -o " + mcl_output_fname + " 2> tmp_mcl.txt";
   // cout << mcl_del << endl << mcl_run << endl;
    if (system(mcl_del.c_str()))  
        cout << "Couldn't delete MCL file";
    if (system(mcl_run.c_str()))
        cout << "Couldn't run MCL";
}

int MclOutput::analyzeClusters(int &number_of_big_clusters, 
    int &number_of_big_good_clusters)
{    
    loadMclCLusters();
    loadMGraph();
    number_of_big_clusters = 0, number_of_big_good_clusters = 0;
    int  contin = 0;
    for (int i = 0; i < n_clusters_mcl; i++)
    {
        int big = isBig(i);
        int good = isGood(i,contin);
        number_of_big_clusters += big;
        if(big && good) 
            number_of_big_good_clusters += 1;        
    }
 
    delete [] mcluster_graph;
    delete [] mcl_cluster; 
    delete [] mgraph;
    return contin;
}

void MclOutput::iterateMcl()
{  
    cout << "Preparing to run MCL" << endl;
    mcl_parameter =  mcl_parameter_start_vlaue;  
    int number_of_big_clusters = 0, number_of_valid_clusters = 0;
    int number_of_big_clusters_prev = 0, number_of_valid_clusters_prev = 0;
    bool repeat = true;
    int maxV = 0;

    while (repeat && mcl_parameter > 1.001)
    {
        cout << "Trying parameter " <<  mcl_parameter << endl;

        runMcl();
        analyzeClusters(number_of_big_clusters, number_of_valid_clusters);

        cout << "Number of big clustrs is " <<  number_of_big_clusters << 
            ". Number of valid clusters is " << number_of_valid_clusters << endl;

        if(number_of_valid_clusters > maxV) 
            maxV = number_of_valid_clusters;

        if ( (number_of_valid_clusters < number_of_valid_clusters_prev 
           && !(number_of_big_clusters >= number_of_big_clusters_prev)) 
             || number_of_big_clusters < maxV) 
            repeat = false; 
        else
        {
            number_of_valid_clusters_prev = number_of_valid_clusters;
            number_of_big_clusters_prev = number_of_big_clusters;
        }

        if (repeat != false)
            mcl_parameter -= 0.01; 
    }  

    mcl_parameter += 0.01;
    cout << "MCL parameter is " <<  mcl_parameter << endl;
    cout << "Writing final M-clusters..." << endl;    
    runMcl();
    analyzeClusters(number_of_big_clusters, number_of_valid_clusters);
    cout << "Number of clusters in the bin is " << number_of_big_clusters << endl;
}

void MclOutput::outputMClusters()
{
    loadMclCLusters();
    loadMGraph();
    ofstream MClusters;
    MClusters.open(final_mclusters.c_str());
    set <int> big_clusters;
    map<int,int> clusters_mapping;
    int bigcluster_counter = 0;

    for (int i = 0; i < n_clusters_mcl; i++)
        if (isBig(i)) 
        {
            bigcluster_counter ++;
            big_clusters.insert(i);
            clusters_mapping[i] = bigcluster_counter;
        }
    
    
    MClusters << bigcluster_counter << " big mclusters" << endl;

    for (int i = 1; i <= n_clusters; i++)
    {
        int mclust = cluster_in_mcluster[i];
        if (big_clusters.count(mclust) > 0)
            MClusters << i << " " << clusters_mapping[mclust] << endl;
        else 
            MClusters << i << " " << -1 << endl;
    }
    MClusters.close(); 
}




	





















