#include <sstream>
#include <iostream>
#include <cstdlib>
#include <string>
#include <sys/stat.h>
#include <sys/types.h>
#include "MainProfile.h"

using namespace std;

void PrintHelp() 
{
  cerr << "Help: TOSS  reads_file  abundancebin_output_prefix  number_of_bins  output_directory" << endl;
}

void ParseOptions(int argc, char **argv, int &l, string &inputfile, string &dir, string & ab_output_prefix, int &ab_n_bins) 
{
    if (argc == 5) 
    {
        l = 20;

        stringstream s_fname(argv[1]);
        if ((s_fname >> inputfile).fail()) 
        {
            cerr << "Error: reads_file is not a string" << endl;
            PrintHelp();
            exit(1);
        }

        stringstream prefix(argv[2]);
        if ((prefix >> ab_output_prefix).fail()) 
        {
            cerr << "Error: abundancebin_output_prefix is not a string" << endl;
            PrintHelp();
            exit(1);
        }

        stringstream n_bins(argv[3]);
        if ((n_bins >> ab_n_bins).fail()) 
        {
            cerr << "Error: number of bins is not an integer" << endl;
            PrintHelp();
            exit(1);
        }

        stringstream s_dir(argv[4]);
        if ((s_dir >> dir).fail()) 
        {
            cerr << "Error: output_directory is not a string" << endl;
            PrintHelp();
            exit(1);
        }
    } 
    else 
    {
        PrintHelp();
        exit(1);
    }
}

void RunToss(int l, string inputfile, string dir, string & ab_output_prefix, int &ab_n_bins)
{
    MainProfile mp(l, inputfile, dir, ab_output_prefix, ab_n_bins);
    mp.checkDir();
    mp.hashLmers();
    mp.processingAbInput();
    for (int i = 0; i < ab_n_bins; i++)
    {
        mp.clusteringBin(i);
        mp.clusterProcessing(i);
    }
    mp.outputBadBins();

}

int main(int argc, char **argv) 
{
    int l, ab_n_bins; 
    string inputfile, dir, ab_output_prefix;

    ParseOptions(argc,argv,l,inputfile,dir, ab_output_prefix,ab_n_bins);
    RunToss(l,inputfile,dir,ab_output_prefix,ab_n_bins);
}


